import asyncio
import aiofile
import aiohttp
import click
import datetime
import hashlib
import json
import logging
import subprocess
import sys
import watchdog

from asyncio import run, gather, wait
from datetime import datetime, timedelta
from grpclib.server import Server
from grpclib.utils import graceful_exit
from hachiko.hachiko import AIOEventHandler, AIOWatchdog
from itertools import chain
from peewee import JOIN
from porerefiner import models
from porerefiner.models import Run, Qa, File, Job, SampleSheet, Sample, Tag, TagJunction
from porerefiner.cli_utils import relativize_path as r, absolutize_path as a, json_formatter
from porerefiner.jobs import poll_jobs, CLASS_REGISTRY, JOBS
from os.path import split, getmtime
from os import remove
from pathlib import Path

from porerefiner.protocols.minknow.rpc.manager_grpc import ManagerServiceStub
from porerefiner.protocols.porerefiner.rpc.porerefiner_pb2 import Run as RunMessage, RunList, RunResponse, Error, GenericResponse, RunRsyncResponse, RunListResponse
from porerefiner.protocols.porerefiner.rpc.porerefiner_grpc import PoreRefinerBase
from porerefiner.notifiers import NOTIFIERS

from porerefiner.rpc import get_run





log = logging.getLogger('porerefiner.fs')



async def register_new_flowcell(flow, nanopore_api=None):
    "Hook for new flowcells"
    log.info(f"Registering flowcell {flow.consumable_id}")
    if nanopore_api:
        pass

async def register_new_run(run, nanopore_api=None):
    "Hook for new runs"
    log.info(f"Registering run {run.name}")
    if nanopore_api:
        pass

    query = list(SampleSheet.get_unused_sheets()) #pre-fetch all
    if len(query) == 1:
        #if there's one unattached sheet, attach it to this run
        sheet = query[0]
        run.sample_sheet = sheet
        run.save()



async def attach_samplesheet_to_run(sheet, run_id=None):
    "Determine file format of sample sheet and load"
    if '.xls' in sheet:
        sheet = await SampleSheet.from_excel(sheet)
    else:
        sheet = await SampleSheet.from_csv(sheet)
    if run_id:
        # run = Run.get_or_none(Run.pk == run_id)
        # if not run:
        #     run = Run.get_or_none(Run.human_name == run_id)
        # if not run:
        #     raise ValueError(f"Run id or name '{run_id}' not found.")
        run = get_run(run_id)
        run.sample_sheet = sheet
        run.save()
    else:
        #find unassociated run
        query = Run.get_unannotated_runs()
        if query.count() == 1:
            run = next(query)
            run.sample_sheet = sheet
            run.save()
        else:
            raise ValueError("Run not specified and there are multiple runs in progress. Please specify a run by id, name, or nickname.")





async def poll_file(file):
    if datetime.now() - file.last_modified > timedelta(hours=1):
        await end_file(file)
        return True
    return False

async def poll_active_run():
    "Scan run(s) in progress, close out runs that have been stable for an hour"
    runs = Run.select().where(Run.status == 'RUNNING')
    i = 0
    for i, run in enumerate(runs, 1):
        if len(run.files) and all([await poll_file(file) for file in run.files]):
            await end_run(run)
    return i


async def end_run(run):
    "Put run in closed status"
    run.ended = datetime.now()
    run.status = 'DONE'
    run.save()
    run.tag('finished')
    log.info(f"Run {run.alt_name} ended, no file modifications in the past hour")
    for notifier in NOTIFIERS:
        log.info(f"Firing notifier {notifier.name}")
        await notifier.notify(run, None, "Run finished")
    for job in JOBS.RUNS:
        log.info(f"Scheduling job {type(job).__name__}")
        run.spawn(job)

async def end_file(file):
    "Put file in closed state"
    log.info(f"No recent modifications to {file.path}, scheduling analysis.")
    # try:
    #     proc = await asyncio.create_subprocess_shell(f'md5sum {file.path}', stdout=subprocess.PIPE)
    #     await proc.wait()
    #     return_val = await proc.stdout.readline()
    #     hash_val = return_val.split(b' ')[0]
    #     file.hash = hash_val
    #     file.save()
    # except (subprocess.CalledProcessError, ValueError) as e:
    #     log.error(e)
    async with aiofile.AIOFile(file.path, 'rb') as afile:
        ha = hashlib.md5()
        rdr = aiofile.Reader(afile, chunk_size=1024 * 1024)
        async for chunk in rdr:
            ha.update(chunk)
    file.hash = ha.hexdigest()
    file.save()
    for job in JOBS.FILES:
        log.info(f"Scheduling job {type(job).__name__} on {file.path}")
        file.spawn(job)


# async def send_run(run, dest):
#     "Use RSYNC to send a run to a destination"
#     pass


class PoreRefinerFSEventHandler(AIOEventHandler):
    "Eventhandler for file system events via Hachiko/Watchdog"

    def __init__(self, path, *a, **k):
        super().__init__(*a, **k)
        self.path = path


    async def on_created(self, event):
        """
        The all-important flowcell, run, and file detection hook.

        RULES: we're watching the nanopore output directory configured by config.

        A folder created under the output directory is an EXPERIMENT.

        A folder created under an experiment is a SAMPLE. We're ignoring both of these because the semantics don't map.

        A folder created under a SAMPLE is a RUN. This, we care about.

        A FILE created anywhere beneath a RUN directory is a file that is part of that run,
        as long as '_porerefiner' isn't part of the path.

        [EXPERIMENT_ID]/[SAMPLE_ID]/[START_TIME]_[DEVICE_ID]_[FLOWCELL_ID]_[SHORT_PROTOCOL_RUN_ID]


        """
        log.info(f"Filesystem event: {event.src_path} created")
        if '_porerefiner' not in str(event.src_path): #may need to exclude analysis results we create ourselves
            path = Path(event.src_path)
            rel = path.relative_to(self.path)
            # if len(rel.parts) >= 1: # not doing flowcells anymore
            #     rel_flow_path, *_ = rel.parts
            #     flow_path = self.path / Path(rel_flow_path)
            #     flow, new = Flowcell.get_or_create(path=r(flow_path), consumable_id=rel_flow_path)
            #     if new:
            #         await register_new_flowcell(flow)
            if len(rel.parts) >= 3:
                exp, sam, rel_run_path, *_ = rel.parts
                run_path = self.path / Path(exp) / Path(sam) / Path(rel_run_path)
                run, new = Run.get_or_create(path=r(run_path), name=rel_run_path)
                if new:
                    run.tag(exp)
                    run.tag(sam)
                    await register_new_run(run)
                    try:
                        st, dev_id, fc_id, prot_id = rel_run_path.split('_')
                        run.flowcell = fc_id
                        # run.tag(st)
                        run.tag(dev_id)
                        run.tag(prot_id)
                    except ValueError:
                        pass
                    run.save()
            if len(rel.parts) >=4 and not event.is_directory: #there's a file
                log.info(f"Registering new file {path} in {run.name}")
                f = File.create(run=run, path=path)
                f.tag(rel.parent.name)
                f.save()




    async def on_modified(self, event):
        if not event.is_directory: #we don't care about directory modifications
            fi = File.get_or_none(File.path == r(event.src_path))
            if not fi:
                await self.on_created(event)
            else:
                fi.last_modified = datetime.now()
                fi.save()

    # async def on_deleted(self, event): #TODO
    #     "Update database if files are deleted."
    #     if event.is_directory:
    #         Run.delete().where(Run.path == r(event.src_path)).execute()
    #     else:
    #         File.delete().where(File.path == r(event.src_path)).execute()


async def start_fs_watchdog(path, api=None, *a, **k):
    "Coroutine to bring up the filesystem watchdog"
    watcher = AIOWatchdog(
        path,
        event_handler=PoreRefinerFSEventHandler(path, *a, **k)
        )
    watcher.start()
    log.info(f"Filesystem events being watched in {path}...")
    # watcher.wait_closed()
    # log.info(f"Filesystem event watcher shutting down.")

async def in_progress_run_update(*args, **kwargs):
    "On server start, update all in-progress run files with last modified date."
    for run in Run.select().where(Run.status == 'RUNNING'):
        log.info(f"Checking in-progress run {run.name} for modifications")
        for file in run.all_files:
            file.last_modified = datetime.fromtimestamp(getmtime(a(file.path)))
            file.save()
            await asyncio.sleep(0)


async def start_run_end_polling(run_polling_interval, *a, **k):
    "Coro to bring up the run termination polling"
    log.info(f"Starting run polling...")
    async def run_end_polling():
        run_num = await poll_active_run()
        log.info(f"{run_num} runs polled.")
        await asyncio.sleep(run_polling_interval) #poll every ten minutes
        return asyncio.ensure_future(run_end_polling())
    return asyncio.ensure_future(run_end_polling())

async def start_job_polling(job_polling_interval, *a, **k):
    log.info(f'Starting job polling...')
    async def run_job_polling():
        po, su, co = await poll_jobs(
            Job.select().where(Job.status == 'READY'),
            Job.select().where(Job.status == 'RUNNING')
        )
        log.info(f'{po} jobs polled, {su} submitted, {co} collected.')
        await asyncio.sleep(job_polling_interval) #poll every 30 minutes
        return asyncio.ensure_future(run_job_polling())
    return asyncio.ensure_future(run_job_polling())



# Generated by the Protocol Buffers compiler. DO NOT EDIT!
# source: minknow/rpc/manager.proto
# plugin: grpclib.plugin.main
import abc
import typing

import grpclib.const
import grpclib.client
if typing.TYPE_CHECKING:
    import grpclib.server

from . import instance_pb2
from . import manager_pb2


class ManagerServiceBase(abc.ABC):

    @abc.abstractmethod
    async def list_devices(self, stream: 'grpclib.server.Stream[minknow.rpc.manager_pb2.ListDevicesRequest, minknow.rpc.manager_pb2.ListDevicesResponse]') -> None:
        pass

    @abc.abstractmethod
    async def get_version_info(self, stream: 'grpclib.server.Stream[minknow.rpc.manager_pb2.GetVersionInfoRequest, minknow.rpc.manager_pb2.GetVersionInfoResponse]') -> None:
        pass

    def __mapping__(self) -> typing.Dict[str, grpclib.const.Handler]:
        return {
            '/ont.rpc.manager.ManagerService/list_devices': grpclib.const.Handler(
                self.list_devices,
                grpclib.const.Cardinality.UNARY_UNARY,
                minknow.rpc.manager_pb2.ListDevicesRequest,
                minknow.rpc.manager_pb2.ListDevicesResponse,
            ),
            '/ont.rpc.manager.ManagerService/get_version_info': grpclib.const.Handler(
                self.get_version_info,
                grpclib.const.Cardinality.UNARY_UNARY,
                minknow.rpc.manager_pb2.GetVersionInfoRequest,
                minknow.rpc.manager_pb2.GetVersionInfoResponse,
            ),
        }


class ManagerServiceStub:

    def __init__(self, channel: grpclib.client.Channel) -> None:
        self.list_devices = grpclib.client.UnaryUnaryMethod(
            channel,
            '/ont.rpc.manager.ManagerService/list_devices',
            minknow.rpc.manager_pb2.ListDevicesRequest,
            minknow.rpc.manager_pb2.ListDevicesResponse,
        )
        self.get_version_info = grpclib.client.UnaryUnaryMethod(
            channel,
            '/ont.rpc.manager.ManagerService/get_version_info',
            minknow.rpc.manager_pb2.GetVersionInfoRequest,
            minknow.rpc.manager_pb2.GetVersionInfoResponse,
        )



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: minknow/rpc/keystore.proto

import sys
_b=sys.version_info[0]<3 and (lambda x:x) or (lambda x:x.encode('latin1'))
from google.protobuf.internal import enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from google.protobuf import any_pb2 as google_dot_protobuf_dot_any__pb2
from minknow.rpc import rpc_options_pb2 as minknow_dot_rpc_dot_rpc__options__pb2


DESCRIPTOR = _descriptor.FileDescriptor(
  name='minknow/rpc/keystore.proto',
  package='ont.rpc.keystore',
  syntax='proto3',
  serialized_options=None,
  serialized_pb=_b('\n\x1aminknow/rpc/keystore.proto\x12\x10ont.rpc.keystore\x1a\x19google/protobuf/any.proto\x1a\x1dminknow/rpc/rpc_options.proto\"\xc3\x01\n\x0cStoreRequest\x12@\n\x06values\x18\x01 \x03(\x0b\x32*.ont.rpc.keystore.StoreRequest.ValuesEntryB\x04\x88\xb5\x18\x01\x12,\n\x08lifetime\x18\x02 \x01(\x0e\x32\x1a.ont.rpc.keystore.Lifetime\x1a\x43\n\x0bValuesEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12#\n\x05value\x18\x02 \x01(\x0b\x32\x14.google.protobuf.Any:\x02\x38\x01\"\x0f\n\rStoreResponse\";\n\rRemoveRequest\x12\x13\n\x05names\x18\x01 \x03(\tB\x04\x88\xb5\x18\x01\x12\x15\n\rallow_missing\x18\x02 \x01(\x08\"\x10\n\x0eRemoveResponse\"#\n\rGetOneRequest\x12\x12\n\x04name\x18\x01 \x01(\tB\x04\x88\xb5\x18\x01\"5\n\x0eGetOneResponse\x12#\n\x05value\x18\x02 \x01(\x0b\x32\x14.google.protobuf.Any\"2\n\nGetRequest\x12\r\n\x05names\x18\x01 \x03(\t\x12\x15\n\rallow_missing\x18\x02 \x01(\x08\"\x8d\x01\n\x0bGetResponse\x12\x39\n\x06values\x18\x01 \x03(\x0b\x32).ont.rpc.keystore.GetResponse.ValuesEntry\x1a\x43\n\x0bValuesEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12#\n\x05value\x18\x02 \x01(\x0b\x32\x14.google.protobuf.Any:\x02\x38\x01\":\n\x0cWatchRequest\x12\x13\n\x05names\x18\x01 \x03(\tB\x04\x88\xb5\x18\x01\x12\x15\n\rallow_missing\x18\x02 \x01(\x08\"\xa9\x01\n\rWatchResponse\x12;\n\x06values\x18\x01 \x03(\x0b\x32+.ont.rpc.keystore.WatchResponse.ValuesEntry\x12\x16\n\x0eremoved_values\x18\x02 \x03(\t\x1a\x43\n\x0bValuesEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12#\n\x05value\x18\x02 \x01(\x0b\x32\x14.google.protobuf.Any:\x02\x38\x01*Y\n\x08Lifetime\x12\x1d\n\x19UNTIL_NEXT_PROTOCOL_START\x10\x00\x12\x16\n\x12UNTIL_PROTOCOL_END\x10\x01\x12\x16\n\x12UNTIL_INSTANCE_END\x10\x02\x32\x90\x03\n\x0fKeyStoreService\x12J\n\x05store\x12\x1e.ont.rpc.keystore.StoreRequest\x1a\x1f.ont.rpc.keystore.StoreResponse\"\x00\x12M\n\x06remove\x12\x1f.ont.rpc.keystore.RemoveRequest\x1a .ont.rpc.keystore.RemoveResponse\"\x00\x12N\n\x07get_one\x12\x1f.ont.rpc.keystore.GetOneRequest\x1a .ont.rpc.keystore.GetOneResponse\"\x00\x12\x44\n\x03get\x12\x1c.ont.rpc.keystore.GetRequest\x1a\x1d.ont.rpc.keystore.GetResponse\"\x00\x12L\n\x05watch\x12\x1e.ont.rpc.keystore.WatchRequest\x1a\x1f.ont.rpc.keystore.WatchResponse\"\x00\x30\x01\x62\x06proto3')
  ,
  dependencies=[google_dot_protobuf_dot_any__pb2.DESCRIPTOR,minknow_dot_rpc_dot_rpc__options__pb2.DESCRIPTOR,])

_LIFETIME = _descriptor.EnumDescriptor(
  name='Lifetime',
  full_name='ont.rpc.keystore.Lifetime',
  filename=None,
  file=DESCRIPTOR,
  values=[
    _descriptor.EnumValueDescriptor(
      name='UNTIL_NEXT_PROTOCOL_START', index=0, number=0,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='UNTIL_PROTOCOL_END', index=1, number=1,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='UNTIL_INSTANCE_END', index=2, number=2,
      serialized_options=None,
      type=None),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=920,
  serialized_end=1009,
)
_sym_db.RegisterEnumDescriptor(_LIFETIME)

Lifetime = enum_type_wrapper.EnumTypeWrapper(_LIFETIME)
UNTIL_NEXT_PROTOCOL_START = 0
UNTIL_PROTOCOL_END = 1
UNTIL_INSTANCE_END = 2



_STOREREQUEST_VALUESENTRY = _descriptor.Descriptor(
  name='ValuesEntry',
  full_name='ont.rpc.keystore.StoreRequest.ValuesEntry',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='key', full_name='ont.rpc.keystore.StoreRequest.ValuesEntry.key', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='value', full_name='ont.rpc.keystore.StoreRequest.ValuesEntry.value', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=_b('8\001'),
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=235,
  serialized_end=302,
)

_STOREREQUEST = _descriptor.Descriptor(
  name='StoreRequest',
  full_name='ont.rpc.keystore.StoreRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='values', full_name='ont.rpc.keystore.StoreRequest.values', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=_b('\210\265\030\001'), file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='lifetime', full_name='ont.rpc.keystore.StoreRequest.lifetime', index=1,
      number=2, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[_STOREREQUEST_VALUESENTRY, ],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=107,
  serialized_end=302,
)


_STORERESPONSE = _descriptor.Descriptor(
  name='StoreResponse',
  full_name='ont.rpc.keystore.StoreResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=304,
  serialized_end=319,
)


_REMOVEREQUEST = _descriptor.Descriptor(
  name='RemoveRequest',
  full_name='ont.rpc.keystore.RemoveRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='names', full_name='ont.rpc.keystore.RemoveRequest.names', index=0,
      number=1, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=_b('\210\265\030\001'), file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='allow_missing', full_name='ont.rpc.keystore.RemoveRequest.allow_missing', index=1,
      number=2, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=321,
  serialized_end=380,
)


_REMOVERESPONSE = _descriptor.Descriptor(
  name='RemoveResponse',
  full_name='ont.rpc.keystore.RemoveResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=382,
  serialized_end=398,
)


_GETONEREQUEST = _descriptor.Descriptor(
  name='GetOneRequest',
  full_name='ont.rpc.keystore.GetOneRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='ont.rpc.keystore.GetOneRequest.name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=_b('\210\265\030\001'), file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=400,
  serialized_end=435,
)


_GETONERESPONSE = _descriptor.Descriptor(
  name='GetOneResponse',
  full_name='ont.rpc.keystore.GetOneResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='value', full_name='ont.rpc.keystore.GetOneResponse.value', index=0,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=437,
  serialized_end=490,
)


_GETREQUEST = _descriptor.Descriptor(
  name='GetRequest',
  full_name='ont.rpc.keystore.GetRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='names', full_name='ont.rpc.keystore.GetRequest.names', index=0,
      number=1, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='allow_missing', full_name='ont.rpc.keystore.GetRequest.allow_missing', index=1,
      number=2, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=492,
  serialized_end=542,
)


_GETRESPONSE_VALUESENTRY = _descriptor.Descriptor(
  name='ValuesEntry',
  full_name='ont.rpc.keystore.GetResponse.ValuesEntry',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='key', full_name='ont.rpc.keystore.GetResponse.ValuesEntry.key', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='value', full_name='ont.rpc.keystore.GetResponse.ValuesEntry.value', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=_b('8\001'),
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=235,
  serialized_end=302,
)

_GETRESPONSE = _descriptor.Descriptor(
  name='GetResponse',
  full_name='ont.rpc.keystore.GetResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='values', full_name='ont.rpc.keystore.GetResponse.values', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[_GETRESPONSE_VALUESENTRY, ],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=545,
  serialized_end=686,
)


_WATCHREQUEST = _descriptor.Descriptor(
  name='WatchRequest',
  full_name='ont.rpc.keystore.WatchRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='names', full_name='ont.rpc.keystore.WatchRequest.names', index=0,
      number=1, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=_b('\210\265\030\001'), file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='allow_missing', full_name='ont.rpc.keystore.WatchRequest.allow_missing', index=1,
      number=2, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=688,
  serialized_end=746,
)


_WATCHRESPONSE_VALUESENTRY = _descriptor.Descriptor(
  name='ValuesEntry',
  full_name='ont.rpc.keystore.WatchResponse.ValuesEntry',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='key', full_name='ont.rpc.keystore.WatchResponse.ValuesEntry.key', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='value', full_name='ont.rpc.keystore.WatchResponse.ValuesEntry.value', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=_b('8\001'),
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=235,
  serialized_end=302,
)

_WATCHRESPONSE = _descriptor.Descriptor(
  name='WatchResponse',
  full_name='ont.rpc.keystore.WatchResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='values', full_name='ont.rpc.keystore.WatchResponse.values', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='removed_values', full_name='ont.rpc.keystore.WatchResponse.removed_values', index=1,
      number=2, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[_WATCHRESPONSE_VALUESENTRY, ],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=749,
  serialized_end=918,
)

_STOREREQUEST_VALUESENTRY.fields_by_name['value'].message_type = google_dot_protobuf_dot_any__pb2._ANY
_STOREREQUEST_VALUESENTRY.containing_type = _STOREREQUEST
_STOREREQUEST.fields_by_name['values'].message_type = _STOREREQUEST_VALUESENTRY
_STOREREQUEST.fields_by_name['lifetime'].enum_type = _LIFETIME
_GETONERESPONSE.fields_by_name['value'].message_type = google_dot_protobuf_dot_any__pb2._ANY
_GETRESPONSE_VALUESENTRY.fields_by_name['value'].message_type = google_dot_protobuf_dot_any__pb2._ANY
_GETRESPONSE_VALUESENTRY.containing_type = _GETRESPONSE
_GETRESPONSE.fields_by_name['values'].message_type = _GETRESPONSE_VALUESENTRY
_WATCHRESPONSE_VALUESENTRY.fields_by_name['value'].message_type = google_dot_protobuf_dot_any__pb2._ANY
_WATCHRESPONSE_VALUESENTRY.containing_type = _WATCHRESPONSE
_WATCHRESPONSE.fields_by_name['values'].message_type = _WATCHRESPONSE_VALUESENTRY
DESCRIPTOR.message_types_by_name['StoreRequest'] = _STOREREQUEST
DESCRIPTOR.message_types_by_name['StoreResponse'] = _STORERESPONSE
DESCRIPTOR.message_types_by_name['RemoveRequest'] = _REMOVEREQUEST
DESCRIPTOR.message_types_by_name['RemoveResponse'] = _REMOVERESPONSE
DESCRIPTOR.message_types_by_name['GetOneRequest'] = _GETONEREQUEST
DESCRIPTOR.message_types_by_name['GetOneResponse'] = _GETONERESPONSE
DESCRIPTOR.message_types_by_name['GetRequest'] = _GETREQUEST
DESCRIPTOR.message_types_by_name['GetResponse'] = _GETRESPONSE
DESCRIPTOR.message_types_by_name['WatchRequest'] = _WATCHREQUEST
DESCRIPTOR.message_types_by_name['WatchResponse'] = _WATCHRESPONSE
DESCRIPTOR.enum_types_by_name['Lifetime'] = _LIFETIME
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

StoreRequest = _reflection.GeneratedProtocolMessageType('StoreRequest', (_message.Message,), {

  'ValuesEntry' : _reflection.GeneratedProtocolMessageType('ValuesEntry', (_message.Message,), {
    'DESCRIPTOR' : _STOREREQUEST_VALUESENTRY,
    '__module__' : 'minknow.rpc.keystore_pb2'
    # @@protoc_insertion_point(class_scope:ont.rpc.keystore.StoreRequest.ValuesEntry)
    })
  ,
  'DESCRIPTOR' : _STOREREQUEST,
  '__module__' : 'minknow.rpc.keystore_pb2'
  # @@protoc_insertion_point(class_scope:ont.rpc.keystore.StoreRequest)
  })
_sym_db.RegisterMessage(StoreRequest)
_sym_db.RegisterMessage(StoreRequest.ValuesEntry)

StoreResponse = _reflection.GeneratedProtocolMessageType('StoreResponse', (_message.Message,), {
  'DESCRIPTOR' : _STORERESPONSE,
  '__module__' : 'minknow.rpc.keystore_pb2'
  # @@protoc_insertion_point(class_scope:ont.rpc.keystore.StoreResponse)
  })
_sym_db.RegisterMessage(StoreResponse)

RemoveRequest = _reflection.GeneratedProtocolMessageType('RemoveRequest', (_message.Message,), {
  'DESCRIPTOR' : _REMOVEREQUEST,
  '__module__' : 'minknow.rpc.keystore_pb2'
  # @@protoc_insertion_point(class_scope:ont.rpc.keystore.RemoveRequest)
  })
_sym_db.RegisterMessage(RemoveRequest)

RemoveResponse = _reflection.GeneratedProtocolMessageType('RemoveResponse', (_message.Message,), {
  'DESCRIPTOR' : _REMOVERESPONSE,
  '__module__' : 'minknow.rpc.keystore_pb2'
  # @@protoc_insertion_point(class_scope:ont.rpc.keystore.RemoveResponse)
  })
_sym_db.RegisterMessage(RemoveResponse)

GetOneRequest = _reflection.GeneratedProtocolMessageType('GetOneRequest', (_message.Message,), {
  'DESCRIPTOR' : _GETONEREQUEST,
  '__module__' : 'minknow.rpc.keystore_pb2'
  # @@protoc_insertion_point(class_scope:ont.rpc.keystore.GetOneRequest)
  })
_sym_db.RegisterMessage(GetOneRequest)

GetOneResponse = _reflection.GeneratedProtocolMessageType('GetOneResponse', (_message.Message,), {
  'DESCRIPTOR' : _GETONERESPONSE,
  '__module__' : 'minknow.rpc.keystore_pb2'
  # @@protoc_insertion_point(class_scope:ont.rpc.keystore.GetOneResponse)
  })
_sym_db.RegisterMessage(GetOneResponse)

GetRequest = _reflection.GeneratedProtocolMessageType('GetRequest', (_message.Message,), {
  'DESCRIPTOR' : _GETREQUEST,
  '__module__' : 'minknow.rpc.keystore_pb2'
  # @@protoc_insertion_point(class_scope:ont.rpc.keystore.GetRequest)
  })
_sym_db.RegisterMessage(GetRequest)

GetResponse = _reflection.GeneratedProtocolMessageType('GetResponse', (_message.Message,), {

  'ValuesEntry' : _reflection.GeneratedProtocolMessageType('ValuesEntry', (_message.Message,), {
    'DESCRIPTOR' : _GETRESPONSE_VALUESENTRY,
    '__module__' : 'minknow.rpc.keystore_pb2'
    # @@protoc_insertion_point(class_scope:ont.rpc.keystore.GetResponse.ValuesEntry)
    })
  ,
  'DESCRIPTOR' : _GETRESPONSE,
  '__module__' : 'minknow.rpc.keystore_pb2'
  # @@protoc_insertion_point(class_scope:ont.rpc.keystore.GetResponse)
  })
_sym_db.RegisterMessage(GetResponse)
_sym_db.RegisterMessage(GetResponse.ValuesEntry)

WatchRequest = _reflection.GeneratedProtocolMessageType('WatchRequest', (_message.Message,), {
  'DESCRIPTOR' : _WATCHREQUEST,
  '__module__' : 'minknow.rpc.keystore_pb2'
  # @@protoc_insertion_point(class_scope:ont.rpc.keystore.WatchRequest)
  })
_sym_db.RegisterMessage(WatchRequest)

WatchResponse = _reflection.GeneratedProtocolMessageType('WatchResponse', (_message.Message,), {

  'ValuesEntry' : _reflection.GeneratedProtocolMessageType('ValuesEntry', (_message.Message,), {
    'DESCRIPTOR' : _WATCHRESPONSE_VALUESENTRY,
    '__module__' : 'minknow.rpc.keystore_pb2'
    # @@protoc_insertion_point(class_scope:ont.rpc.keystore.WatchResponse.ValuesEntry)
    })
  ,
  'DESCRIPTOR' : _WATCHRESPONSE,
  '__module__' : 'minknow.rpc.keystore_pb2'
  # @@protoc_insertion_point(class_scope:ont.rpc.keystore.WatchResponse)
  })
_sym_db.RegisterMessage(WatchResponse)
_sym_db.RegisterMessage(WatchResponse.ValuesEntry)


_STOREREQUEST_VALUESENTRY._options = None
_STOREREQUEST.fields_by_name['values']._options = None
_REMOVEREQUEST.fields_by_name['names']._options = None
_GETONEREQUEST.fields_by_name['name']._options = None
_GETRESPONSE_VALUESENTRY._options = None
_WATCHREQUEST.fields_by_name['names']._options = None
_WATCHRESPONSE_VALUESENTRY._options = None

_KEYSTORESERVICE = _descriptor.ServiceDescriptor(
  name='KeyStoreService',
  full_name='ont.rpc.keystore.KeyStoreService',
  file=DESCRIPTOR,
  index=0,
  serialized_options=None,
  serialized_start=1012,
  serialized_end=1412,
  methods=[
  _descriptor.MethodDescriptor(
    name='store',
    full_name='ont.rpc.keystore.KeyStoreService.store',
    index=0,
    containing_service=None,
    input_type=_STOREREQUEST,
    output_type=_STORERESPONSE,
    serialized_options=None,
  ),
  _descriptor.MethodDescriptor(
    name='remove',
    full_name='ont.rpc.keystore.KeyStoreService.remove',
    index=1,
    containing_service=None,
    input_type=_REMOVEREQUEST,
    output_type=_REMOVERESPONSE,
    serialized_options=None,
  ),
  _descriptor.MethodDescriptor(
    name='get_one',
    full_name='ont.rpc.keystore.KeyStoreService.get_one',
    index=2,
    containing_service=None,
    input_type=_GETONEREQUEST,
    output_type=_GETONERESPONSE,
    serialized_options=None,
  ),
  _descriptor.MethodDescriptor(
    name='get',
    full_name='ont.rpc.keystore.KeyStoreService.get',
    index=3,
    containing_service=None,
    input_type=_GETREQUEST,
    output_type=_GETRESPONSE,
    serialized_options=None,
  ),
  _descriptor.MethodDescriptor(
    name='watch',
    full_name='ont.rpc.keystore.KeyStoreService.watch',
    index=4,
    containing_service=None,
    input_type=_WATCHREQUEST,
    output_type=_WATCHRESPONSE,
    serialized_options=None,
  ),
])
_sym_db.RegisterServiceDescriptor(_KEYSTORESERVICE)

DESCRIPTOR.services_by_name['KeyStoreService'] = _KEYSTORESERVICE

# @@protoc_insertion_point(module_scope)



from flask import Flask, request
import os
from pathlib import Path
from subprocess import run

app = application = Flask(__name__)
app.config['config_file'] = os.environ.get('POREREFINER_CONFIG', Path.home() / '.porerefiner' / 'config.yaml')
app.config['host'] = os.environ.get('HOSTNAME', run(["hostname"], capture_output=True, text=True).stdout.strip())

from . import main
from . import page



# Generated by the Protocol Buffers compiler. DO NOT EDIT!
# source: minknow/rpc/acquisition.proto
# plugin: grpclib.plugin.main
import abc
import typing

import grpclib.const
import grpclib.client
if typing.TYPE_CHECKING:
    import grpclib.server

import google.protobuf.timestamp_pb2
from . import acquisition_pb2


class AcquisitionServiceBase(abc.ABC):

    @abc.abstractmethod
    async def start(self, stream: 'grpclib.server.Stream[minknow.rpc.acquisition_pb2.StartRequest, minknow.rpc.acquisition_pb2.StartResponse]') -> None:
        pass

    @abc.abstractmethod
    async def stop(self, stream: 'grpclib.server.Stream[minknow.rpc.acquisition_pb2.StopRequest, minknow.rpc.acquisition_pb2.StopResponse]') -> None:
        pass

    @abc.abstractmethod
    async def watch_for_status_change(self, stream: 'grpclib.server.Stream[minknow.rpc.acquisition_pb2.WatchForStatusChangeRequest, minknow.rpc.acquisition_pb2.WatchForStatusChangeResponse]') -> None:
        pass

    @abc.abstractmethod
    async def watch_current_acquisition_run(self, stream: 'grpclib.server.Stream[minknow.rpc.acquisition_pb2.WatchCurrentAcquisitionRunRequest, minknow.rpc.acquisition_pb2.AcquisitionRunInfo]') -> None:
        pass

    @abc.abstractmethod
    async def current_status(self, stream: 'grpclib.server.Stream[minknow.rpc.acquisition_pb2.CurrentStatusRequest, minknow.rpc.acquisition_pb2.CurrentStatusResponse]') -> None:
        pass

    @abc.abstractmethod
    async def get_progress(self, stream: 'grpclib.server.Stream[minknow.rpc.acquisition_pb2.GetProgressRequest, minknow.rpc.acquisition_pb2.GetProgressResponse]') -> None:
        pass

    @abc.abstractmethod
    async def get_acquisition_info(self, stream: 'grpclib.server.Stream[minknow.rpc.acquisition_pb2.GetAcquisitionRunInfoRequest, minknow.rpc.acquisition_pb2.AcquisitionRunInfo]') -> None:
        pass

    @abc.abstractmethod
    async def list_acquisition_runs(self, stream: 'grpclib.server.Stream[minknow.rpc.acquisition_pb2.ListAcquisitionRunsRequest, minknow.rpc.acquisition_pb2.ListAcquisitionRunsResponse]') -> None:
        pass

    @abc.abstractmethod
    async def get_current_acquisition_run(self, stream: 'grpclib.server.Stream[minknow.rpc.acquisition_pb2.GetCurrentAcquisitionRunRequest, minknow.rpc.acquisition_pb2.AcquisitionRunInfo]') -> None:
        pass

    def __mapping__(self) -> typing.Dict[str, grpclib.const.Handler]:
        return {
            '/ont.rpc.acquisition.AcquisitionService/start': grpclib.const.Handler(
                self.start,
                grpclib.const.Cardinality.UNARY_UNARY,
                minknow.rpc.acquisition_pb2.StartRequest,
                minknow.rpc.acquisition_pb2.StartResponse,
            ),
            '/ont.rpc.acquisition.AcquisitionService/stop': grpclib.const.Handler(
                self.stop,
                grpclib.const.Cardinality.UNARY_UNARY,
                minknow.rpc.acquisition_pb2.StopRequest,
                minknow.rpc.acquisition_pb2.StopResponse,
            ),
            '/ont.rpc.acquisition.AcquisitionService/watch_for_status_change': grpclib.const.Handler(
                self.watch_for_status_change,
                grpclib.const.Cardinality.STREAM_STREAM,
                minknow.rpc.acquisition_pb2.WatchForStatusChangeRequest,
                minknow.rpc.acquisition_pb2.WatchForStatusChangeResponse,
            ),
            '/ont.rpc.acquisition.AcquisitionService/watch_current_acquisition_run': grpclib.const.Handler(
                self.watch_current_acquisition_run,
                grpclib.const.Cardinality.UNARY_STREAM,
                minknow.rpc.acquisition_pb2.WatchCurrentAcquisitionRunRequest,
                minknow.rpc.acquisition_pb2.AcquisitionRunInfo,
            ),
            '/ont.rpc.acquisition.AcquisitionService/current_status': grpclib.const.Handler(
                self.current_status,
                grpclib.const.Cardinality.UNARY_UNARY,
                minknow.rpc.acquisition_pb2.CurrentStatusRequest,
                minknow.rpc.acquisition_pb2.CurrentStatusResponse,
            ),
            '/ont.rpc.acquisition.AcquisitionService/get_progress': grpclib.const.Handler(
                self.get_progress,
                grpclib.const.Cardinality.UNARY_UNARY,
                minknow.rpc.acquisition_pb2.GetProgressRequest,
                minknow.rpc.acquisition_pb2.GetProgressResponse,
            ),
            '/ont.rpc.acquisition.AcquisitionService/get_acquisition_info': grpclib.const.Handler(
                self.get_acquisition_info,
                grpclib.const.Cardinality.UNARY_UNARY,
                minknow.rpc.acquisition_pb2.GetAcquisitionRunInfoRequest,
                minknow.rpc.acquisition_pb2.AcquisitionRunInfo,
            ),
            '/ont.rpc.acquisition.AcquisitionService/list_acquisition_runs': grpclib.const.Handler(
                self.list_acquisition_runs,
                grpclib.const.Cardinality.UNARY_UNARY,
                minknow.rpc.acquisition_pb2.ListAcquisitionRunsRequest,
                minknow.rpc.acquisition_pb2.ListAcquisitionRunsResponse,
            ),
            '/ont.rpc.acquisition.AcquisitionService/get_current_acquisition_run': grpclib.const.Handler(
                self.get_current_acquisition_run,
                grpclib.const.Cardinality.UNARY_UNARY,
                minknow.rpc.acquisition_pb2.GetCurrentAcquisitionRunRequest,
                minknow.rpc.acquisition_pb2.AcquisitionRunInfo,
            ),
        }


class AcquisitionServiceStub:

    def __init__(self, channel: grpclib.client.Channel) -> None:
        self.start = grpclib.client.UnaryUnaryMethod(
            channel,
            '/ont.rpc.acquisition.AcquisitionService/start',
            minknow.rpc.acquisition_pb2.StartRequest,
            minknow.rpc.acquisition_pb2.StartResponse,
        )
        self.stop = grpclib.client.UnaryUnaryMethod(
            channel,
            '/ont.rpc.acquisition.AcquisitionService/stop',
            minknow.rpc.acquisition_pb2.StopRequest,
            minknow.rpc.acquisition_pb2.StopResponse,
        )
        self.watch_for_status_change = grpclib.client.StreamStreamMethod(
            channel,
            '/ont.rpc.acquisition.AcquisitionService/watch_for_status_change',
            minknow.rpc.acquisition_pb2.WatchForStatusChangeRequest,
            minknow.rpc.acquisition_pb2.WatchForStatusChangeResponse,
        )
        self.watch_current_acquisition_run = grpclib.client.UnaryStreamMethod(
            channel,
            '/ont.rpc.acquisition.AcquisitionService/watch_current_acquisition_run',
            minknow.rpc.acquisition_pb2.WatchCurrentAcquisitionRunRequest,
            minknow.rpc.acquisition_pb2.AcquisitionRunInfo,
        )
        self.current_status = grpclib.client.UnaryUnaryMethod(
            channel,
            '/ont.rpc.acquisition.AcquisitionService/current_status',
            minknow.rpc.acquisition_pb2.CurrentStatusRequest,
            minknow.rpc.acquisition_pb2.CurrentStatusResponse,
        )
        self.get_progress = grpclib.client.UnaryUnaryMethod(
            channel,
            '/ont.rpc.acquisition.AcquisitionService/get_progress',
            minknow.rpc.acquisition_pb2.GetProgressRequest,
            minknow.rpc.acquisition_pb2.GetProgressResponse,
        )
        self.get_acquisition_info = grpclib.client.UnaryUnaryMethod(
            channel,
            '/ont.rpc.acquisition.AcquisitionService/get_acquisition_info',
            minknow.rpc.acquisition_pb2.GetAcquisitionRunInfoRequest,
            minknow.rpc.acquisition_pb2.AcquisitionRunInfo,
        )
        self.list_acquisition_runs = grpclib.client.UnaryUnaryMethod(
            channel,
            '/ont.rpc.acquisition.AcquisitionService/list_acquisition_runs',
            minknow.rpc.acquisition_pb2.ListAcquisitionRunsRequest,
            minknow.rpc.acquisition_pb2.ListAcquisitionRunsResponse,
        )
        self.get_current_acquisition_run = grpclib.client.UnaryUnaryMethod(
            channel,
            '/ont.rpc.acquisition.AcquisitionService/get_current_acquisition_run',
            minknow.rpc.acquisition_pb2.GetCurrentAcquisitionRunRequest,
            minknow.rpc.acquisition_pb2.AcquisitionRunInfo,
        )



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Tests for `porerefiner` package."""

from unittest import TestCase, skip
from unittest.mock import Mock, patch
#from unittest.mock import AsyncMock
from mock import AsyncMock
from aiounittest import async_test

from click.testing import CliRunner

from porerefiner import porerefiner, fsevents as pr_fsevents, rpc
from porerefiner import cli
from porerefiner import models
from porerefiner import config
from porerefiner.cli_utils import absolutize_path as ap, relativize_path as rp
from porerefiner.protocols.porerefiner.rpc import porerefiner_pb2 as messages
from tests import paths, with_database, TestBase as DBSetupTestCase, fsevents, Model

from shutil import rmtree

from os.path import split

from tempfile import mktemp, mkdtemp, mkstemp, TemporaryDirectory, NamedTemporaryFile

from hypothesis import given, settings, HealthCheck
import hypothesis.strategies as strat
#from hypothesis_fspaths import fspaths

import asyncio
#from asyncio import run as _run

import pathlib
import os, sys

from datetime import datetime, timedelta

from peewee import JOIN

def _run(task):
    loop = asyncio.get_event_loop()
    return loop.run_until_complete(task)


RUN_PK = 101
RUN_NAME = 'TEST_TEST'

class TestCoreFunctions(DBSetupTestCase):

    def setUp(self):
        super().setUp()
        # self.flow = flow = models.Flowcell.create(consumable_id="TEST|TEST|TEST", consumable_type="TEST|TEST|TEST", path="TEST/TEST/TEST")
        self._tempdir = t = TemporaryDirectory()
        self._testdir = t = pathlib.Path(t.name)
        self.run = r = models.Run.create(pk=RUN_PK, library_id='x', name=RUN_NAME, path=t)
        _, t = mkstemp(dir=t)
        self.file = f = models.File.create(run=r, path=pathlib.Path(t), last_modified=datetime.now() - timedelta(hours=2))

    def tearDown(self):
        self._tempdir.cleanup()



    def test_get_run(self):
        run1 = rpc.get_run(RUN_PK)
        run2 = rpc.get_run(RUN_NAME)
        assert run1 == run2

    # @given(fspaths())
    # def test_fs_paths(self, path):
    #     pathlib.Path(bytes.decode(path))


    @given(strat.one_of(strat.text(max_size=30), strat.integers(min_value=-2**16, max_value=2**16)))
    def test_fail_get_run(self, runid):
        with self.assertRaises(ValueError):
            run = rpc.get_run(runid)


    def test_get_run_info(self):
        run1 = _run(rpc.get_run_info(RUN_PK))
        run2 = _run(rpc.get_run_info(RUN_NAME))
        self.assertEqual(run1, run2)

    @given(strat.one_of(
            strat.text(max_size=30).filter(lambda n: n != RUN_NAME),
            strat.integers(min_value=-2**16, max_value=2**16).filter(lambda n: n != RUN_PK)
            ))
    def test_fail_get_run_info(self, run_id):
        with self.assertRaises(ValueError):
            _run(rpc.get_run_info(run_id))

    #@skip('not implemented')
    def test_list_runs(self):
        self.assertEqual(len(_run(rpc.list_runs(all=True))), 1)

    def test_list_runs_running(self):
        models.Run.create(library_id='x', name=RUN_NAME, path="TEST/TEST/TEST", ended=datetime.now())
        self.assertEqual(len(_run(rpc.list_runs())), 1)

    def test_list_runs_tags(self):
        tag = models.Tag.create(name='TEST')
        models.TagJunction.create(tag=tag, run=self.run)
        self.assertEqual(len(_run(rpc.list_runs(tags=['TEST', 'other tag']))), 1)


    #@skip('not implemented')
    @patch('porerefiner.fsevents.end_run', new_callable=AsyncMock)
    def test_poll_active_run(self, mock):
        #run = models.Run.get(models.Run.pk==RUN_PK)
        self.assertGreater(len(models.Run.select().where(models.Run.status == 'RUNNING')), 0)
        self.assertGreater(len(models.Run.get(models.Run.pk==RUN_PK).files), 0)
        self.assertEqual(_run(pr_fsevents.poll_active_run()), 1) #checked 1 file
        #self.assertGreater(len(run.files), 0)
        #self.assertGreater(len(list(models.Run.select(models.Run.ended.is_null(True)))), 0)
        #self.assertTrue(len(run.files) and all([datetime.now() - file.last_modified > timedelta(hours=1) for file in run.files]))
        mock.assert_called() #ran end_run


    @skip('cant make this work with the database')
    @with_database
    def test_end_run(self):
        mock = AsyncMock()
        with patch('porerefiner.fsevents.NOTIFIERS', new_callable=lambda: [mock]) as _:
            _run(pr_fsevents.end_run(self.run))
        self.assertAlmostEqual(datetime.now(), self.run.ended, delta=timedelta(seconds=1))
        mock.notify.assert_called()
        self.assertIn('finished', [str(tag) for tag in self.run.tags])

    # @skip('not implemented')
    # def test_send_run(self):
    #     assert False

    @patch('porerefiner.fsevents.SampleSheet')
    @patch('porerefiner.fsevents.logging')
    def test_register_new_run(self, log, mock):
        mock_run = Mock()
        mock.get_unused_sheets.return_value = [None, ] #just need one value
        self.assertGreater(len(mock.get_unused_sheets()), 0)
        self.assertIs(pr_fsevents.SampleSheet, mock)
        _run(pr_fsevents.register_new_run(mock_run))
        mock_run.save.assert_called()





# @patch('porerefiner.porerefiner.register_new_run', new_callable=AsyncMock)
# @patch('porerefiner.porerefiner.register_new_flowcell', new_callable=AsyncMock)
# @patch('porerefiner.porerefiner.Flowcell')
# @patch('porerefiner.porerefiner.Run')
# @patch('porerefiner.porerefiner.File')
# class TestPoreFSEventHander(TestCase):

#     class FakeEvent:
#         def __init__(self, path, is_dir=True):
#             self.src_path = path
#             self.is_directory = is_dir



#     @given(event=fsevents())
#     def test_on_created(self, event, *a, **k):
#         ut = porerefiner.PoreRefinerFSEventHandler(event.src_path.parent)
#         _run(ut.on_created(event))
#         assert True


#     #@skip('not implemented')

#     @given(flowcell_path=paths().filter(lambda p: isinstance(p, pathlib.Path)))
#     def test_on_created_flowcell(self, flowcell_path, *a, **k):
#         ut = porerefiner.PoreRefinerFSEventHandler(flowcell_path.parent)
#         event = self.FakeEvent(flowcell_path)
#         _run(ut.on_created(event))


#     def test_on_created_run_simple(self, path=pathlib.Path('/A/B/C/D/E'), *a, **k):
#         ut = porerefiner.PoreRefinerFSEventHandler(path.parent.parent)
#         event = self.FakeEvent(path)
#         _run(ut.on_created(event))


#     def test_on_created_run_deep(self, path=pathlib.Path('/A/B/C/D/E'), *a, **k):
#         ut = porerefiner.PoreRefinerFSEventHandler(path.parent.parent.parent)
#         event = self.FakeEvent(path)
#         _run(ut.on_created(event))



#     # @skip('not implemented')
#     # @with_database
#     @patch('porerefiner.porerefiner.Flowcell')
#     def test_on_created_file(self, mock_flow, path=pathlib.Path('/A/B/C/D/E'), *a, **k):
#         ut = porerefiner.PoreRefinerFSEventHandler(path.parent.parent.parent)
#         mock_flow.get_or_create.return_value = mock_run.get_or_create.return_value = (Mock(), False)
#         event = self.FakeEvent(rp(path), is_dir=False)
#         _run(ut.on_created(event))


#     def test_on_created_file_deep(self, path=pathlib.Path('/A/B/C/D/E'), *a, **k):
#         ut = porerefiner.PoreRefinerFSEventHandler(path.parent.parent.parent) #A/B
#         event = self.FakeEvent(path / 'test', is_dir=False) #E/test
#         _run(ut.on_created(event))


#     # @skip('not implemented')
#     @patch('porerefiner.porerefiner.File')
#     def test_on_modified(self, mock):
#         ut = porerefiner.PoreRefinerFSEventHandler(pathlib.Path('TEST'))
#         event = self.FakeEvent('TEST', False)
#         _run(ut.on_modified(event))
#         mock.get_or_none.assert_called()

#     @skip('not implemented')
#     @given(paths())
#     def test_on_deleted(self, path):
#         assert False

class TestPoreDispatchServer(TestCase):

    # def setUp(self):
    #     super().setUp()
    #     self.flow = flow = models.Flowcell.create(consumable_id="TEST|TEST|TEST", consumable_type="TEST|TEST|TEST", path="TEST/TEST/TEST")
    #     self.run = models.Run.create(pk=RUN_PK, library_id='x', name=RUN_NAME, flowcell=flow, path="TEST/TEST/TEST")
    #     self.file = models.File.create(run=self.run, path='TEST/TEST/TEST/TEST', last_modified=datetime.now() - timedelta(hours=2))
    #     self.tag = models.Tag.create(name='TEST')
    #     models.TagJunction.create(tag=tag, run=self.run)

    # @skip('no test')
    @settings(deadline=500, suppress_health_check=(HealthCheck.all()))
    @given(ss=Model.Samplesheets())
    @with_database
    def test_attach_sheet_run_no_run(self, ss):
        ut = rpc.PoreRefinerDispatchServer()
        strm = AsyncMock()
        strm.recv_message.return_value = messages.RunAttachRequest(sheet=ss)
        _run(ut.AttachSheetToRun(strm))
        strm.send_message.assert_called_once()

    @settings(deadline=500, suppress_health_check=(HealthCheck.all()))
    @given(ss=Model.Samplesheets())
    @with_database
    def test_attach_sheet_to_run(self, ss):
        # self.flow = flow = models.Flowcell.create(consumable_id="TEST|TEST|TEST", consumable_type="TEST|TEST|TEST", path="TEST/TEST/TEST")
        self.run = models.Run.create(pk=RUN_PK, library_id='x', name=RUN_NAME, path="TEST/TEST/TEST")
        self.file = models.File.create(run=self.run, path='TEST/TEST/TEST/TEST', last_modified=datetime.now() - timedelta(hours=2))
        ut = rpc.PoreRefinerDispatchServer()
        strm = AsyncMock()
        strm.recv_message.return_value = messages.RunAttachRequest(sheet=ss, id=RUN_PK)
        _run(ut.AttachSheetToRun(strm))
        strm.send_message.assert_called_once()



class TestServerStart(TestCase):

    #@skip('no test')
    @patch('porerefiner.rpc.graceful_exit')
    @patch('porerefiner.rpc.Server')
    def test_start_server(self, mock, _):
        mock.return_value = coro = AsyncMock()
        _run(rpc.start_server(None))
        mock.assert_called()
        coro.start.assert_called()
        coro.wait_closed.assert_called()

    #@skip('no test')
    @with_database
    @async_test
    @patch('porerefiner.fsevents.logging')
    async def test_start_run_end_polling(self, log):
        with patch('porerefiner.fsevents.poll_active_run') as mock:
            mock.return_value = 1
            task = await pr_fsevents.start_run_end_polling(0)
            await asyncio.sleep(5)
            task.cancel()
            mock.assert_called()

    #@skip('no test')
    @with_database
    @async_test
    async def test_start_job_polling(self):
        with patch('porerefiner.fsevents.poll_jobs') as mock:
            task = await pr_fsevents.start_job_polling(0)
            await asyncio.sleep(5)
            task.cancel()
            mock.assert_called()

class TestConfig(TestCase):

    def test_config_create_defaults(self):
        f = mktemp()
        c = config.Config(f)
        self.assertTrue(c.config)
        os.unlink(f)




# -*- coding: utf-8 -*-

"""Unit test package for porerefiner."""

# from hypothesis_fspaths import fspaths

#safe_paths = lambda: fspaths().filter(lambda x: isinstance(x, str) or hasattr(x, '__fspath__'))

import asyncio
import tempfile
import logging
import sys

from unittest import TestCase

from collections import namedtuple
from datetime import datetime
from hypothesis.strategies import *
from pathlib import Path
import os
import namesgenerator

from porerefiner import models, jobs

from peewee import SqliteDatabase

from porerefiner.protocols.porerefiner.rpc import porerefiner_pb2 as messages

the_present = datetime.now()

logging.basicConfig(stream=sys.stdout,
                    style='{',
                    format="{levelname} {name}:{message}",
                    level=logging.ERROR)

# SQLite can't accept a 32-bit integer
sql_ints = lambda: integers(min_value=-2**16, max_value=2**16)

def _run(task):
    loop = asyncio.get_event_loop()
    return loop.run_until_complete(task)

@composite
def paths(draw, under="", min_deep=2, max_deep=None, pathlib_only=False):
    if max_deep and max_deep < min_deep:
        raise ValueError("max_deep must be greater than min_deep")
    r = [just(under)]
    for _ in range(min_deep):
        r.append(text(min_size=3, max_size=255))
    if max_deep:
        import random
        for _ in range(random.randint(0, max_deep-min_deep)):
            r.append(text(min_size=3, max_size=255))
    p = builds(Path, *r)
    if pathlib_only:
        return draw(p)
    return draw(one_of(p, p.map(str)))

# @composite
# def files(draw, real=False):
#     if real:
#         path = builds(Path, builds(lambda *a, **k: tempfile.NamedTemporaryFile().name))
#     else:
#         path = paths(pathlib_only=True)
#     return draw(builds(models.File,
#                        pk=sql_ints(),
#                        path=path))


# @composite
# def samples(draw):
#     sid = text(min_size=7, max_size=10)
#     acc = text(min_size=5, max_size=10)
#     bar = text(min_size=10, max_size=10)
#     org = text(min_size=12, max_size=12)
#     ext = text(min_size=10, max_size=10)
#     com = text()
#     use = emails()
#     return draw(builds(messages.SampleSheet.Sample,
#                        sample_id=sid,
#                        accession=acc,
#                        barcode_id=bar,
#                        organism=org,
#                        extraction_kit=ext,
#                        comment=com,
#                        user=use))

# @composite
# def samplesheets(draw):
#     ver = just('1.0.1')
#     dat = datetimes()
#     lib = text(min_size=12, max_size=12)
#     seq = text(min_size=12, max_size=12)
#     bar = lists(text(min_size=12, max_size=12), min_size=0, max_size=3)
#     sam = lists(samples(), min_size=1, max_size=12)
#     ss = draw(builds(messages.SampleSheet,
#                      porerefiner_ver=ver,
#                      library_id=lib,
#                      sequencing_kit=seq,
#                      barcode_kit=bar,
#                      samples=sam))
#     ss.date.FromDatetime(draw(dat))
#     return ss


# @composite
# def flowcells(draw, path=None):
#     pk  = sql_ints()
#     cid = text()
#     cname = text()
#     cty = text()
#     if not path:
#         pat = paths(min_deep=2, under="/data")
#     else:
#         pat = just(Path(path))
#     return draw(builds(models.Flowcell,
#                        #pk=pk,
#                        consumable_id=cid,
#                        consumable_name=cname,
#                        consumable_type=cty,
#                        path=pat))

# @composite
# def runs(draw, sheet=True):
#     if sheet:
#         ss = samplesheets()
#     else:
#         ss = just(None)
#     path = draw(paths(pathlib_only=True))
#     return draw(builds(models.Run,
#                        #pk=sql_ints(),
#                        # flowcell=flowcells(path=path.parent),
#                        _sample_sheets=ss,
#                        name=text(),
#                        library_id=text(),
#                        run_id=text(),
#                        started=datetimes(max_value=the_present),
#                        ended=datetimes(min_value=the_present),
#                        status=sampled_from([status[0] for status in models.Run.statuses]),
#                        path=just(path),
#                        basecalling_model=sampled_from([model[0] for model in models.Run.basecallers])))

Event = namedtuple('Event', ('src_path', 'is_directory'))

@composite
def fsevents(draw, min_deep=4, isDirectory=None):
    if isDirectory is None:
        is_dir = booleans()
    else:
        is_dir = just(bool(isDirectory))
    return draw(builds(Event,
                       src_path=paths(min_deep=min_deep, pathlib_only=True),
                       is_directory=is_dir))

def random_name_subclass(of=object, **classdef):
    classdef['__module__'] = __name__
    new_typename = namesgenerator.get_random_name(sep=' ').title().replace(' ', '') + of.__name__
    new_type = type(new_typename, (of,), classdef)
    globals()[new_typename] = new_type
    return new_type

class TestSubmitter(jobs.submitters.Submitter):
    async def test_noop(*a, **k):
        return None
    def reroot_path(*a, **k):
        return Path()
    async def begin_job(*a, **k):
        return "fake started"
    async def poll_job(*a, **k):
        return "fake status"
    def closeout_job(*a, **k):
        return None

@composite
def submitters(draw):
    # async def fake_noop(*a, **k):
    #     return None
    # async def fake_begin(*a, **k):
    #     return "fake started"
    # async def fake_poll(*a, **k):
    #     return "fake status"
    # return draw(builds(random_name_subclass(of=subclass_of,
    #                                         test_noop=fake_noop,
    #                                         reroot_path=lambda *a, **k: Path(),
    #                                         begin_job=fake_begin,
    #                                         poll_job=fake_poll,
    #                                         closeout_job=lambda *a, **k: None)))
    return draw(just(TestSubmitter()))

class TestJob(jobs.RunJob):

    def setup(*a, **k):
        return "", {}
    def collect(*a, **k):
        return None

@composite
def jobs(draw):
    # return draw(builds(random_name_subclass(of=subclass_of, **classdef),
    #                    submitter=submitters()))
    return draw(just(TestJob(submitter=TestSubmitter())))



# @composite
# def job_records(draw, state=jobs()):
#     return draw(builds(models.Job,
#                        job_state=state))

class Model:

    @staticmethod
    @composite
    def Jobs(draw, state=jobs()):
        TestJob(TestSubmitter())
        return draw(builds(models.Job,
                       job_class=just(TestJob.__name__),
                       datadir=just("/dev/null")))

    @staticmethod
    @composite
    def Runs(draw, sheet=True):
        if sheet:
            ss = Model.Samplesheets()
        else:
            ss = just(None)
        path = draw(paths(pathlib_only=True))
        return draw(builds(models.Run,
                        #pk=sql_ints(),
                        # flowcell=flowcells(path=path.parent),
                        _sample_sheets=ss,
                        name=text(),
                        library_id=text(),
                        run_id=text(),
                        started=datetimes(max_value=the_present),
                        ended=datetimes(min_value=the_present),
                        status=sampled_from([status[0] for status in models.Run.statuses]),
                        path=just(path),
                        basecalling_model=sampled_from([model[0] for model in models.Run.basecallers])))

    @staticmethod
    @composite
    def Files(draw, real=False):
        if real:
            path = builds(Path, builds(lambda *a, **k: tempfile.NamedTemporaryFile().name))
        else:
            path = paths(pathlib_only=True)
        return draw(builds(models.File,
                        pk=sql_ints(),
                        path=path))

    @staticmethod
    @composite
    def Samples(draw):
        sid = text(min_size=7, max_size=10)
        acc = text(min_size=5, max_size=10)
        bar = text(min_size=10, max_size=10)
        org = text(min_size=12, max_size=12)
        ext = text(min_size=10, max_size=10)
        com = text()
        use = emails()
        return draw(builds(messages.SampleSheet.Sample,
                        sample_id=sid,
                        accession=acc,
                        barcode_id=bar,
                        organism=org,
                        extraction_kit=ext,
                        comment=com,
                        user=use))

    @staticmethod
    @composite
    def Samplesheets(draw):
        ver = just('1.0.1')
        dat = datetimes()
        lib = text(min_size=12, max_size=12)
        seq = text(min_size=12, max_size=12)
        bar = lists(text(min_size=12, max_size=12), min_size=0, max_size=3)
        sam = lists(Model.Samples(), min_size=1, max_size=12)
        ss = draw(builds(messages.SampleSheet,
                        porerefiner_ver=ver,
                        library_id=lib,
                        sequencing_kit=seq,
                        barcode_kit=bar,
                        samples=sam))
        ss.date.FromDatetime(draw(dat))
        return ss

    @staticmethod
    @composite
    def Flowcells(draw, path=None):
        pk  = sql_ints()
        cid = text()
        cname = text()
        cty = text()
        if not path:
            pat = paths(min_deep=2, under="/data")
        else:
            pat = just(Path(path))
        return draw(builds(models.Flowcell,
                        #pk=pk,
                        consumable_id=cid,
                        consumable_name=cname,
                        consumable_type=cty,
                        path=pat))

class TestBase(TestCase):

    def setUp(self):
        self.db = SqliteDatabase(":memory:", pragmas={'foreign_keys':1}, autoconnect=False)
        self.db.bind(models.REGISTRY, bind_refs=False, bind_backrefs=False)
        self.db.connect()
        self.db.create_tables(models.REGISTRY)

    def tearDown(self):
        #[cls.delete().where(True).execute() for cls in models.REGISTRY]
        self.db.drop_tables(models.REGISTRY)
        self.db.close()

from functools import wraps

def with_database(func):
    @wraps(func)
    def wrapped_test_function(*a, **k):
        db = SqliteDatabase(":memory:", pragmas={'foreign_keys':1}, autoconnect=False)
        db.bind(models.REGISTRY, bind_refs=True, bind_backrefs=True)
        db.connect()
        db.create_tables(models.REGISTRY)
        try:
            return func(*a, **k)
        finally:
            db.drop_tables(models.REGISTRY)
            db.close()
    return wrapped_test_function

if __name__ == '__main__':
    symbol = None
    for symbol in locals().values():
        if hasattr(symbol, 'is_hypothesis_strategy_function'):
            try:
                print(symbol, symbol().example())
            except:
                pass



import asyncio
import logging
import json

from dataclasses import dataclass

from porerefiner.notifiers import Notifier

log = logging.getLogger('porerefiner.sqs_notifier')

@dataclass
class SqsNotifier(Notifier): #TODO
    "Send a message to an SQS queue, requires Boto3"

    queue: str = "porerefiner-messages"


    async def notify(self, run, state, message):
        try:
            import boto3
            sqs = boto3.resource('sqs')
            q = sqs.get_queue_by_name(QueueName=self.queue)
            attrs = dict(State=dict(StringVal=str(state), DataType='String'),
                         Message=dict(StringVal=str(message), DataType='String'))
            response = q.send_message(MessageBody=json.dumps(run), MessageAttributes=attrs)
            log.info("SQS mesage sent.")

        except ImportError:
            log.error("SQS notifier enabled, but Boto3 not installed.")
        except Exception as e:
            log.error(f"Error sending SQS message: {e}")



from tests import _run
from tests import *

from unittest import TestCase

import porerefiner.jobs.submitters as subs

from hypothesis import given

class TestSubmitters(TestCase):

    @given(job_rec = Model.Jobs(),
           job_code = jobs())
    @with_database
    def test_submit(self, job_rec, job_code):
        _run(job_code.submitter._submit(job_rec))


# Generated by the Protocol Buffers compiler. DO NOT EDIT!
# source: minknow/rpc/device.proto
# plugin: grpclib.plugin.main
import abc
import typing

import grpclib.const
import grpclib.client
if typing.TYPE_CHECKING:
    import grpclib.server

from . import rpc_options_pb2
import google.protobuf.wrappers_pb2
from . import device_pb2


class DeviceServiceBase(abc.ABC):

    @abc.abstractmethod
    async def get_device_info(self, stream: 'grpclib.server.Stream[minknow.rpc.device_pb2.GetDeviceInfoRequest, minknow.rpc.device_pb2.GetDeviceInfoResponse]') -> None:
        pass

    @abc.abstractmethod
    async def get_device_state(self, stream: 'grpclib.server.Stream[minknow.rpc.device_pb2.GetDeviceStateRequest, minknow.rpc.device_pb2.GetDeviceStateResponse]') -> None:
        pass

    @abc.abstractmethod
    async def stream_device_state(self, stream: 'grpclib.server.Stream[minknow.rpc.device_pb2.StreamDeviceStateRequest, minknow.rpc.device_pb2.GetDeviceStateResponse]') -> None:
        pass

    @abc.abstractmethod
    async def get_flow_cell_info(self, stream: 'grpclib.server.Stream[minknow.rpc.device_pb2.GetFlowCellInfoRequest, minknow.rpc.device_pb2.GetFlowCellInfoResponse]') -> None:
        pass

    @abc.abstractmethod
    async def stream_flow_cell_info(self, stream: 'grpclib.server.Stream[minknow.rpc.device_pb2.StreamFlowCellInfoRequest, minknow.rpc.device_pb2.GetFlowCellInfoResponse]') -> None:
        pass

    @abc.abstractmethod
    async def set_user_specified_flow_cell_id(self, stream: 'grpclib.server.Stream[minknow.rpc.device_pb2.SetUserSpecifiedFlowCellIdRequest, minknow.rpc.device_pb2.SetUserSpecifiedFlowCellIdResponse]') -> None:
        pass

    @abc.abstractmethod
    async def set_user_specified_product_code(self, stream: 'grpclib.server.Stream[minknow.rpc.device_pb2.SetUserSpecifiedProductCodeRequest, minknow.rpc.device_pb2.SetUserSpecifiedProductCodeResponse]') -> None:
        pass

    @abc.abstractmethod
    async def get_channels_layout(self, stream: 'grpclib.server.Stream[minknow.rpc.device_pb2.GetChannelsLayoutRequest, minknow.rpc.device_pb2.GetChannelsLayoutResponse]') -> None:
        pass

    def __mapping__(self) -> typing.Dict[str, grpclib.const.Handler]:
        return {
            '/ont.rpc.device.DeviceService/get_device_info': grpclib.const.Handler(
                self.get_device_info,
                grpclib.const.Cardinality.UNARY_UNARY,
                minknow.rpc.device_pb2.GetDeviceInfoRequest,
                minknow.rpc.device_pb2.GetDeviceInfoResponse,
            ),
            '/ont.rpc.device.DeviceService/get_device_state': grpclib.const.Handler(
                self.get_device_state,
                grpclib.const.Cardinality.UNARY_UNARY,
                minknow.rpc.device_pb2.GetDeviceStateRequest,
                minknow.rpc.device_pb2.GetDeviceStateResponse,
            ),
            '/ont.rpc.device.DeviceService/stream_device_state': grpclib.const.Handler(
                self.stream_device_state,
                grpclib.const.Cardinality.UNARY_STREAM,
                minknow.rpc.device_pb2.StreamDeviceStateRequest,
                minknow.rpc.device_pb2.GetDeviceStateResponse,
            ),
            '/ont.rpc.device.DeviceService/get_flow_cell_info': grpclib.const.Handler(
                self.get_flow_cell_info,
                grpclib.const.Cardinality.UNARY_UNARY,
                minknow.rpc.device_pb2.GetFlowCellInfoRequest,
                minknow.rpc.device_pb2.GetFlowCellInfoResponse,
            ),
            '/ont.rpc.device.DeviceService/stream_flow_cell_info': grpclib.const.Handler(
                self.stream_flow_cell_info,
                grpclib.const.Cardinality.UNARY_STREAM,
                minknow.rpc.device_pb2.StreamFlowCellInfoRequest,
                minknow.rpc.device_pb2.GetFlowCellInfoResponse,
            ),
            '/ont.rpc.device.DeviceService/set_user_specified_flow_cell_id': grpclib.const.Handler(
                self.set_user_specified_flow_cell_id,
                grpclib.const.Cardinality.UNARY_UNARY,
                minknow.rpc.device_pb2.SetUserSpecifiedFlowCellIdRequest,
                minknow.rpc.device_pb2.SetUserSpecifiedFlowCellIdResponse,
            ),
            '/ont.rpc.device.DeviceService/set_user_specified_product_code': grpclib.const.Handler(
                self.set_user_specified_product_code,
                grpclib.const.Cardinality.UNARY_UNARY,
                minknow.rpc.device_pb2.SetUserSpecifiedProductCodeRequest,
                minknow.rpc.device_pb2.SetUserSpecifiedProductCodeResponse,
            ),
            '/ont.rpc.device.DeviceService/get_channels_layout': grpclib.const.Handler(
                self.get_channels_layout,
                grpclib.const.Cardinality.UNARY_UNARY,
                minknow.rpc.device_pb2.GetChannelsLayoutRequest,
                minknow.rpc.device_pb2.GetChannelsLayoutResponse,
            ),
        }


class DeviceServiceStub:

    def __init__(self, channel: grpclib.client.Channel) -> None:
        self.get_device_info = grpclib.client.UnaryUnaryMethod(
            channel,
            '/ont.rpc.device.DeviceService/get_device_info',
            minknow.rpc.device_pb2.GetDeviceInfoRequest,
            minknow.rpc.device_pb2.GetDeviceInfoResponse,
        )
        self.get_device_state = grpclib.client.UnaryUnaryMethod(
            channel,
            '/ont.rpc.device.DeviceService/get_device_state',
            minknow.rpc.device_pb2.GetDeviceStateRequest,
            minknow.rpc.device_pb2.GetDeviceStateResponse,
        )
        self.stream_device_state = grpclib.client.UnaryStreamMethod(
            channel,
            '/ont.rpc.device.DeviceService/stream_device_state',
            minknow.rpc.device_pb2.StreamDeviceStateRequest,
            minknow.rpc.device_pb2.GetDeviceStateResponse,
        )
        self.get_flow_cell_info = grpclib.client.UnaryUnaryMethod(
            channel,
            '/ont.rpc.device.DeviceService/get_flow_cell_info',
            minknow.rpc.device_pb2.GetFlowCellInfoRequest,
            minknow.rpc.device_pb2.GetFlowCellInfoResponse,
        )
        self.stream_flow_cell_info = grpclib.client.UnaryStreamMethod(
            channel,
            '/ont.rpc.device.DeviceService/stream_flow_cell_info',
            minknow.rpc.device_pb2.StreamFlowCellInfoRequest,
            minknow.rpc.device_pb2.GetFlowCellInfoResponse,
        )
        self.set_user_specified_flow_cell_id = grpclib.client.UnaryUnaryMethod(
            channel,
            '/ont.rpc.device.DeviceService/set_user_specified_flow_cell_id',
            minknow.rpc.device_pb2.SetUserSpecifiedFlowCellIdRequest,
            minknow.rpc.device_pb2.SetUserSpecifiedFlowCellIdResponse,
        )
        self.set_user_specified_product_code = grpclib.client.UnaryUnaryMethod(
            channel,
            '/ont.rpc.device.DeviceService/set_user_specified_product_code',
            minknow.rpc.device_pb2.SetUserSpecifiedProductCodeRequest,
            minknow.rpc.device_pb2.SetUserSpecifiedProductCodeResponse,
        )
        self.get_channels_layout = grpclib.client.UnaryUnaryMethod(
            channel,
            '/ont.rpc.device.DeviceService/get_channels_layout',
            minknow.rpc.device_pb2.GetChannelsLayoutRequest,
            minknow.rpc.device_pb2.GetChannelsLayoutResponse,
        )



# Generated by the Protocol Buffers compiler. DO NOT EDIT!
# source: minknow/rpc/instance.proto
# plugin: grpclib.plugin.main
import abc
import typing

import grpclib.const
import grpclib.client
if typing.TYPE_CHECKING:
    import grpclib.server

from . import acquisition_pb2
from . import device_pb2
from . import protocol_pb2
from . import instance_pb2


class InstanceServiceBase(abc.ABC):

    @abc.abstractmethod
    async def get_version_info(self, stream: 'grpclib.server.Stream[minknow.rpc.instance_pb2.GetVersionInfoRequest, minknow.rpc.instance_pb2.GetVersionInfoResponse]') -> None:
        pass

    @abc.abstractmethod
    async def get_output_directories(self, stream: 'grpclib.server.Stream[minknow.rpc.instance_pb2.GetOutputDirectoriesRequest, minknow.rpc.instance_pb2.OutputDirectories]') -> None:
        pass

    @abc.abstractmethod
    async def get_default_output_directories(self, stream: 'grpclib.server.Stream[minknow.rpc.instance_pb2.GetDefaultOutputDirectoriesRequest, minknow.rpc.instance_pb2.OutputDirectories]') -> None:
        pass

    @abc.abstractmethod
    async def set_output_directory(self, stream: 'grpclib.server.Stream[minknow.rpc.instance_pb2.SetOutputDirectoryRequest, minknow.rpc.instance_pb2.SetOutputDirectoryResponse]') -> None:
        pass

    @abc.abstractmethod
    async def set_reads_directory(self, stream: 'grpclib.server.Stream[minknow.rpc.instance_pb2.SetReadsDirectoryRequest, minknow.rpc.instance_pb2.SetReadsDirectoryResponse]') -> None:
        pass

    @abc.abstractmethod
    async def get_disk_space_info(self, stream: 'grpclib.server.Stream[minknow.rpc.instance_pb2.GetDiskSpaceInfoRequest, minknow.rpc.instance_pb2.GetDiskSpaceInfoResponse]') -> None:
        pass

    @abc.abstractmethod
    async def get_machine_id(self, stream: 'grpclib.server.Stream[minknow.rpc.instance_pb2.GetMachineIdRequest, minknow.rpc.instance_pb2.GetMachineIdResponse]') -> None:
        pass

    @abc.abstractmethod
    async def get_host_type(self, stream: 'grpclib.server.Stream[minknow.rpc.instance_pb2.GetHostTypeRequest, minknow.rpc.instance_pb2.GetHostTypeResponse]') -> None:
        pass

    @abc.abstractmethod
    async def stream_instance_activity(self, stream: 'grpclib.server.Stream[minknow.rpc.instance_pb2.StreamInstanceActivityRequest, minknow.rpc.instance_pb2.StreamInstanceActivityResponse]') -> None:
        pass

    def __mapping__(self) -> typing.Dict[str, grpclib.const.Handler]:
        return {
            '/ont.rpc.instance.InstanceService/get_version_info': grpclib.const.Handler(
                self.get_version_info,
                grpclib.const.Cardinality.UNARY_UNARY,
                minknow.rpc.instance_pb2.GetVersionInfoRequest,
                minknow.rpc.instance_pb2.GetVersionInfoResponse,
            ),
            '/ont.rpc.instance.InstanceService/get_output_directories': grpclib.const.Handler(
                self.get_output_directories,
                grpclib.const.Cardinality.UNARY_UNARY,
                minknow.rpc.instance_pb2.GetOutputDirectoriesRequest,
                minknow.rpc.instance_pb2.OutputDirectories,
            ),
            '/ont.rpc.instance.InstanceService/get_default_output_directories': grpclib.const.Handler(
                self.get_default_output_directories,
                grpclib.const.Cardinality.UNARY_UNARY,
                minknow.rpc.instance_pb2.GetDefaultOutputDirectoriesRequest,
                minknow.rpc.instance_pb2.OutputDirectories,
            ),
            '/ont.rpc.instance.InstanceService/set_output_directory': grpclib.const.Handler(
                self.set_output_directory,
                grpclib.const.Cardinality.UNARY_UNARY,
                minknow.rpc.instance_pb2.SetOutputDirectoryRequest,
                minknow.rpc.instance_pb2.SetOutputDirectoryResponse,
            ),
            '/ont.rpc.instance.InstanceService/set_reads_directory': grpclib.const.Handler(
                self.set_reads_directory,
                grpclib.const.Cardinality.UNARY_UNARY,
                minknow.rpc.instance_pb2.SetReadsDirectoryRequest,
                minknow.rpc.instance_pb2.SetReadsDirectoryResponse,
            ),
            '/ont.rpc.instance.InstanceService/get_disk_space_info': grpclib.const.Handler(
                self.get_disk_space_info,
                grpclib.const.Cardinality.UNARY_UNARY,
                minknow.rpc.instance_pb2.GetDiskSpaceInfoRequest,
                minknow.rpc.instance_pb2.GetDiskSpaceInfoResponse,
            ),
            '/ont.rpc.instance.InstanceService/get_machine_id': grpclib.const.Handler(
                self.get_machine_id,
                grpclib.const.Cardinality.UNARY_UNARY,
                minknow.rpc.instance_pb2.GetMachineIdRequest,
                minknow.rpc.instance_pb2.GetMachineIdResponse,
            ),
            '/ont.rpc.instance.InstanceService/get_host_type': grpclib.const.Handler(
                self.get_host_type,
                grpclib.const.Cardinality.UNARY_UNARY,
                minknow.rpc.instance_pb2.GetHostTypeRequest,
                minknow.rpc.instance_pb2.GetHostTypeResponse,
            ),
            '/ont.rpc.instance.InstanceService/stream_instance_activity': grpclib.const.Handler(
                self.stream_instance_activity,
                grpclib.const.Cardinality.UNARY_STREAM,
                minknow.rpc.instance_pb2.StreamInstanceActivityRequest,
                minknow.rpc.instance_pb2.StreamInstanceActivityResponse,
            ),
        }


class InstanceServiceStub:

    def __init__(self, channel: grpclib.client.Channel) -> None:
        self.get_version_info = grpclib.client.UnaryUnaryMethod(
            channel,
            '/ont.rpc.instance.InstanceService/get_version_info',
            minknow.rpc.instance_pb2.GetVersionInfoRequest,
            minknow.rpc.instance_pb2.GetVersionInfoResponse,
        )
        self.get_output_directories = grpclib.client.UnaryUnaryMethod(
            channel,
            '/ont.rpc.instance.InstanceService/get_output_directories',
            minknow.rpc.instance_pb2.GetOutputDirectoriesRequest,
            minknow.rpc.instance_pb2.OutputDirectories,
        )
        self.get_default_output_directories = grpclib.client.UnaryUnaryMethod(
            channel,
            '/ont.rpc.instance.InstanceService/get_default_output_directories',
            minknow.rpc.instance_pb2.GetDefaultOutputDirectoriesRequest,
            minknow.rpc.instance_pb2.OutputDirectories,
        )
        self.set_output_directory = grpclib.client.UnaryUnaryMethod(
            channel,
            '/ont.rpc.instance.InstanceService/set_output_directory',
            minknow.rpc.instance_pb2.SetOutputDirectoryRequest,
            minknow.rpc.instance_pb2.SetOutputDirectoryResponse,
        )
        self.set_reads_directory = grpclib.client.UnaryUnaryMethod(
            channel,
            '/ont.rpc.instance.InstanceService/set_reads_directory',
            minknow.rpc.instance_pb2.SetReadsDirectoryRequest,
            minknow.rpc.instance_pb2.SetReadsDirectoryResponse,
        )
        self.get_disk_space_info = grpclib.client.UnaryUnaryMethod(
            channel,
            '/ont.rpc.instance.InstanceService/get_disk_space_info',
            minknow.rpc.instance_pb2.GetDiskSpaceInfoRequest,
            minknow.rpc.instance_pb2.GetDiskSpaceInfoResponse,
        )
        self.get_machine_id = grpclib.client.UnaryUnaryMethod(
            channel,
            '/ont.rpc.instance.InstanceService/get_machine_id',
            minknow.rpc.instance_pb2.GetMachineIdRequest,
            minknow.rpc.instance_pb2.GetMachineIdResponse,
        )
        self.get_host_type = grpclib.client.UnaryUnaryMethod(
            channel,
            '/ont.rpc.instance.InstanceService/get_host_type',
            minknow.rpc.instance_pb2.GetHostTypeRequest,
            minknow.rpc.instance_pb2.GetHostTypeResponse,
        )
        self.stream_instance_activity = grpclib.client.UnaryStreamMethod(
            channel,
            '/ont.rpc.instance.InstanceService/stream_instance_activity',
            minknow.rpc.instance_pb2.StreamInstanceActivityRequest,
            minknow.rpc.instance_pb2.StreamInstanceActivityResponse,
        )



from abc import ABCMeta, abstractmethod, ABC
from asyncio import gather
from collections import namedtuple
from dataclasses import dataclass

from .submitters import Submitter
from porerefiner.models import Job, File, Run
from porerefiner.cli_utils import render_dataclass, Email, Url, PathStr
from pathlib import Path
from typing import Union, Tuple

import logging
import pkgutil
import traceback

JOBS = namedtuple('JOBS', ('FILES', 'RUNS'), defaults=([], []))() #configured (reified) job instances

CLASS_REGISTRY = {} # available job classes

CONFIGURED_JOB_REGISTRY = {}

log = logging.getLogger('porerefiner.job')

async def poll_active_job(job):
    logg = log.getChild(f"poll")
    configured_job = CONFIGURED_JOB_REGISTRY[job.job_class]
    submitter = configured_job.submitter
    try:
        await submitter._poll(job)
    except Exception as e:
        logg.error(f"error in {type(submitter).__name__} {type(configured_job).__name__} ")
        # logg.error(e)
        # traceback.print_exc()
        raise
    return 1

async def submit_job(job):
    logg = log.getChild(f"submit")
    configured_job = CONFIGURED_JOB_REGISTRY[job.job_class]
    submitter = configured_job.submitter
    try:
        await submitter._submit(job)
    except Exception as e:
        logg.error(f"error in {type(submitter).__name__} {type(configured_job).__name__} ")
        # logg.error(e)
        # traceback.print_exc()
        raise
    return 1

async def complete_job(job):
    job.job_state.submitter._close(job)
    return 1

async def poll_jobs(ready_jobs, running_jobs):
    # jobs_submitted = 0
    # jobs_collected = 0
    # for job in Job.select().where(Job.status == 'READY'):
    #     jobs_submitted += await submit_job(job)
    #     jobs_polled += 1
    jobs_submitted = sum(await gather(*[submit_job(job) for job in ready_jobs]))
    # for job in Job.select().where(Job.status == 'RUNNING'):
    #     jobs_collected += await poll_active_job(job)
    #     jobs_polled += 1
    jobs_collected = sum(await gather(*[poll_active_job(job) for job in running_jobs]))
    return jobs_submitted + jobs_collected, jobs_submitted, jobs_collected




class _MetaRegistry(type):

    def __new__(meta, name, bases, class_dict):
        "Register new CLASSES of job"
        cls = type.__new__(meta, name, bases, class_dict)
        if cls not in CLASS_REGISTRY:
            CLASS_REGISTRY[name] = cls
        return cls

    # def __call__(cls, *args, **kwargs):
    #     try:
    #         the_instance = super().__call__(*args, **kwargs)
    #     except TypeError as e:
    #         raise TypeError(f"Bad config: {cls.__name__}: {e} ({args}, {kwargs})") from e
    #     if isinstance(the_instance, FileJob):
    #         JOBS.FILES.append(the_instance)
    #         log.getChild('files').debug(cls.__name__)
    #     if isinstance(the_instance, RunJob):
    #         JOBS.RUNS.append(the_instance)
    #         log.getChild('runs').debug(cls.__name__)
    #     return the_instance
    def __call__(cls, *args, **kwargs):
        "Register new INSTANCES of jobs"
        try:
            the_instance = super().__call__(*args, **kwargs)
        except TypeError as e:
            raise TypeError(f"Bad config: {cls.__name__}: {e} ({args}, {kwargs})") from e
        except:
            raise
        if isinstance(the_instance, FileJob):
            JOBS.FILES.append(the_instance)
            log.getChild('files').debug(cls.__name__)
        if isinstance(the_instance, RunJob):
            JOBS.RUNS.append(the_instance)
            log.getChild('runs').debug(cls.__name__)
        CONFIGURED_JOB_REGISTRY[cls.__name__] = the_instance
        assert the_instance
        return the_instance

class RegisteringABCMeta(ABCMeta, _MetaRegistry):

    pass


@dataclass
class AbstractJob(metaclass=RegisteringABCMeta):

    submitter: Submitter

    @classmethod
    def get_configurable_options(cls):
        "Enumerate configurable options and value type as a guide to configuration."
        return render_dataclass(cls)


class FileJob(AbstractJob):

    @abstractmethod
    def setup(self, run: Run, file: File, datadir: Path, remotedir: Path) -> Union[str, Tuple[str, dict]]:
        pass

    @abstractmethod
    def collect(self, run: Run, file: File, datadir: Path, pid: Union[str, int]) -> None:
        pass

class RunJob(AbstractJob):

    @abstractmethod
    def setup(self, run: Run, datadir: Path, remotedir: Path) -> Union[str, Tuple[str, dict]]:
        pass

    @abstractmethod
    def collect(self, run: Run, datadir: Path, pid: Union[str, int]) -> None:
        pass



for loader, module_name, is_pkg in  pkgutil.walk_packages(__path__):
    _module = loader.find_module(module_name).load_module(module_name)
    # globals()[module_name] = _module



from tests import _run, fsevents, with_database

from unittest import TestCase, skip
from unittest.mock import Mock, patch, AsyncMock
# from mock import AsyncMock
from tempfile import NamedTemporaryFile

from porerefiner.fsevents import PoreRefinerFSEventHandler as Handler, end_file, end_run, register_new_run
from porerefiner.models import Run

from hypothesis import given, settings, note
import hypothesis.strategies as strat

class TestPoreRefinerFSEventsHandler(TestCase):

    @skip('no flowcell support anymore')
    # @patch('porerefiner.fsevents.register_new_flowcell', new_callable=AsyncMock)
    # @patch('porerefiner.fsevents.Flowcell')
    # @given(event=fsevents().filter(lambda e: e.is_directory))
    def test_on_created_flowcell(self, event, reg):
        # flow.get_or_create.return_value = (flow, True)
        _run(Handler(event.src_path.parent).on_created(event))
        reg.assert_called()

    # @skip("can't debug test")
    @patch('porerefiner.fsevents.logging')
    @given(event=fsevents(min_deep=4, isDirectory=True).filter(lambda e: len(e.src_path.parts) > 3)) #three levels deepl plus one
    @with_database
    def test_on_created_run(self, log, event):
        with patch('porerefiner.fsevents.Run') as run:
            self.assertGreater(len(event.src_path.parts), 2)
            note(event.src_path)
            run.get_or_create.return_value = (run, True)
            _run(Handler(event.src_path.parts[0]).on_created(event))
            run.save.assert_called()

    # @skip("can't debug test")
    # @patch('porerefiner.fsevents.Flowcell')
    @settings(deadline=500)
    @given(event=fsevents(min_deep=5, isDirectory=False).filter(lambda e: len(e.src_path.parts) > 4)) # run, plus one more for file
    @with_database
    def test_on_created_file(self, event):
        with patch('porerefiner.fsevents.Run') as run:
            with patch('porerefiner.fsevents.File') as file:
                self.assertGreater(len(event.src_path.parts), 3)
                note(event.src_path)
                run.get_or_create.return_value = (run, False)
                _run(Handler(event.src_path.parts[0]).on_created(event))
                file.create.assert_called()

class TestFSEventsPollingFunctions(TestCase):

    #@given(file=files(real=True))
    def test_end_file(self, file=Mock()):
        with NamedTemporaryFile() as tfile:
            file.path = tfile.name
            _run(end_file(file))
            file.save.assert_called()
